#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdbool.h>

#include "adv_config.h"
#include "retained.h"
#include "main.h"
#include <zephyr/kernel.h>
#include <zephyr/sys/reboot.h>

void setConfigBit(uint16_t *dest, int bitNumber, int val) {
    if(val > 0) {
        *dest |= (1U << bitNumber);
    } else if (val == 0) {
        *dest &= ~(1U << bitNumber);
    }
}

void configureSensor(uint8_t sensor, uint16_t time){
    setConfigBit(&retained.sensor_select,sensor,time);
    retained.sensor_time_config[sensor]=time;
}


int parse_config(char buf[]) {
    char *start = buf;
    char *end = buf;
    
    // Define the allowed first words
    char *allowedFirstWords[] = {"timeout", "wakeup", "sensor", "sleep", "lpcomp", "debug", "cal", "accel", "ts", "reset", "baud"};

    // Find the first word
    char *firstWord = NULL;
    for (int i = 0; i < sizeof(allowedFirstWords) / sizeof(allowedFirstWords[0]); i++) {
        if (strstr(start, allowedFirstWords[i]) == start) {
            firstWord = allowedFirstWords[i];
            break;
        }
    }

    if (firstWord == NULL) {        
        return SBOX_CONFIG_PARSE_ERROR_UNKNOWN_PARAM;
    }

    // Skip the first word
    start += strlen(firstWord);
    // Find the second word
    while (*start == ' ' || *start == '\t' || *start == '\n') {
        start++;
    }
    end = start;
    while (*end != ' ' && *end != '\t' && *end != '\n' && *end != '\0') {
        end++;
    }
    
    char secondWordBuf[16];
    int32_t integerValue;
    uint32_t ts;
    snprintf(secondWordBuf, sizeof(secondWordBuf), "%.*s", (int)(end - start), start);

    if (strcmp(firstWord, "timeout") == 0) {        
        if(sscanf(secondWordBuf,"%d", &integerValue) == 1) {
            retained.timeout=integerValue;
        } else {
            return SBOX_CONFIG_PARSE_ERROR_INVALID_ARG;
        }  
    } else if (strcmp(firstWord, "sleep") == 0) {        
        if(sscanf(secondWordBuf,"%d", &integerValue) == 1) {
            retained.sleep_loop=integerValue;
        } else {
            return SBOX_CONFIG_PARSE_ERROR_INVALID_ARG;
        }
    } else if (strcmp(firstWord, "debug") == 0) {      
        if(sscanf(secondWordBuf,"%d", &integerValue) == 1) {
            retained.debug=integerValue;
        } else {
            return SBOX_CONFIG_PARSE_ERROR_INVALID_ARG;
        }    
    } else if (strcmp(firstWord, "lpcomp") == 0) {      
        if(sscanf(secondWordBuf,"%d", &integerValue) == 1) {
            retained.lpcomp_setup=integerValue;
        } else {
            return SBOX_CONFIG_PARSE_ERROR_INVALID_ARG;
        }
    } else if (strcmp(firstWord, "cal") == 0) {
        calibrate_analog_inputs();
    } else if (strcmp(firstWord, "accel") == 0) {
        if(sscanf(secondWordBuf,"%d", &integerValue) == 1) {
            retained.accel_sensitivity=integerValue;
            start = end;
            while (*start == ' ' || *start == '\t' || *start == '\n') {
                start++;
            }
            end = start;
            while (*end != ' ' && *end != '\t' && *end != '\n' && *end != '\0') {
                end++;
            }
            snprintf(secondWordBuf,sizeof(secondWordBuf),"%.*s",(int)(end - start),start);
            if(sscanf(secondWordBuf,"%d", &integerValue) == 1) {
                retained.accel_freq = integerValue;
            } else {
                return SBOX_CONFIG_PARSE_ERROR_INVALID_ARG;    
            }
        } else {
            return SBOX_CONFIG_PARSE_ERROR_INVALID_ARG;
        }
    } else if(strcmp(firstWord, "ts") == 0) {
        if(sscanf(secondWordBuf,"%d", &ts) == 1) {
            retained.config_timestamp=ts;
        } else {
            return SBOX_CONFIG_PARSE_ERROR_INVALID_ARG;
        }  
    } else if(strcmp(firstWord, "reset") == 0) {        
        sys_reboot(SYS_REBOOT_COLD);
    } else if(strcmp(firstWord, "baud") == 0) {
        if(sscanf(secondWordBuf,"%d", &integerValue) == 1) {
            retained.uart_baud=integerValue;
        } else {
            return SBOX_CONFIG_PARSE_ERROR_INVALID_ARG;
        }
    } else if (strcmp(firstWord, "sensor") == 0) {
        // Skip the space
        start = end;
        // Find and print the optional integer value
        while (*start == ' ' || *start == '\t' || *start == '\n') {
            start++;
        }    
        if (sscanf(start, "%d", &integerValue) == 1) {
            // OK
        } else {
            return SBOX_CONFIG_PARSE_ERROR_INVALID_ARG;
        }
        if(strcmp(secondWordBuf, "accel") == 0) {
            configureSensor(SBOX_CONFIG_SENSOR_ACCEL, integerValue);
        } else if(strcmp(secondWordBuf, "gyro") == 0) {
            configureSensor(SBOX_CONFIG_SENSOR_GYRO, integerValue);
        } else if(strcmp(secondWordBuf, "analog") == 0) {
            configureSensor(SBOX_CONFIG_SENSOR_ANALOG_IN, integerValue);        
        } else if(strcmp(secondWordBuf, "counter") == 0) {
            configureSensor(SBOX_CONFIG_SENSOR_TILT_COUNTER, integerValue);        
        } else if(strcmp(secondWordBuf, "rs232") == 0) {
            configureSensor(SBOX_CONFIG_SENSOR_RS232, integerValue);
        } else if(strcmp(secondWordBuf, "rs485") == 0) {
            configureSensor(SBOX_CONFIG_SENSOR_RS485, integerValue);        
        } else if(strcmp(secondWordBuf, "1wire") == 0) {
            configureSensor(SBOX_CONFIG_SENSOR_1WIRE, integerValue);                  
        } else {
            return SBOX_CONFIG_PARSE_ERROR_UNKNOWN_PARAM;
        }
    } 
    if (strcmp(firstWord, "wakeup") == 0) {        
        if(strcmp(secondWordBuf, "accel") == 0) {
            setConfigBit(&retained.wakeup_source, SBOX_CONFIG_WAKEUP_ACCEL, integerValue);
        } else if(strcmp(secondWordBuf, "tilt") == 0) {
            setConfigBit(&retained.wakeup_source, SBOX_CONFIG_WAKEUP_TILT_COUNTER, integerValue);    
        } else if(strcmp(secondWordBuf, "counter") == 0) {
            setConfigBit(&retained.wakeup_source, SBOX_CONFIG_WAKEUP_IN_COUNTER, integerValue);        
        } else if(strcmp(secondWordBuf, "onoff") == 0) {
            setConfigBit(&retained.wakeup_source, SBOX_CONFIG_WAKEUP_ONOFF, integerValue);
        } else if(strcmp(secondWordBuf, "rs232") == 0) {
            setConfigBit(&retained.wakeup_source, SBOX_CONFIG_WAKEUP_RS232, integerValue);
        } else if(strcmp(secondWordBuf, "rs485") == 0) {
            setConfigBit(&retained.wakeup_source, SBOX_CONFIG_WAKEUP_RS485, integerValue);        
        } else if(strcmp(secondWordBuf, "1wire") == 0) {
            setConfigBit(&retained.wakeup_source, SBOX_CONFIG_WAKEUP_1WIRE, integerValue);
        } else if(strcmp(secondWordBuf, "analog_in") == 0) {
            setConfigBit(&retained.wakeup_source, SBOX_CONFIG_WAKEUP_ANALOG_IN, integerValue);
        } else if(strcmp(secondWordBuf, "dc_in") == 0) {
            setConfigBit(&retained.wakeup_source, SBOX_CONFIG_WAKEUP_DC_IN, integerValue);    
        } else if(strcmp(secondWordBuf, "20ma") == 0) {
            setConfigBit(&retained.wakeup_source, SBOX_CONFIG_WAKEUP_20MA, integerValue);
        } else {
            return SBOX_CONFIG_PARSE_ERROR_UNKNOWN_PARAM;    
        }        
    }

    return SBOX_CONFIG_PARSE_OK;
}
