/*
 * Copyright (c) 2021 Nordic Semiconductor ASA
 *
 * SPDX-License-Identifier: Apache-2.0
 */

#ifndef RETAINED_H_
#define RETAINED_H_

#ifdef __cplusplus
extern "C" {
#endif

#include <inttypes.h>

/* Example of validatable retained data. */
struct retained_data {
	/* The uptime from the current session the last time the
	 * retained data was updated.
	 */
	uint64_t uptime_latest;

	/* Cumulative uptime from all previous sessions up through
	 * uptime_latest of this session.
	 */
	uint64_t uptime_sum;

	/* Number of times the application has started. */
	uint32_t boots;

	/* Number of times the application has gone into system off. */
	uint32_t off_count;

	/* Timeout until go to power_off: positive value - timeout counter in sleep_loop s, negative: never go to power_off*/
	int16_t timeout;
	uint16_t sleep_loop; // main loop sleep in ms
	uint16_t wakeup_source; // also reset timeout
	uint16_t sensor_select;
	uint16_t sensor_time_config[16];
	uint16_t debug;
	uint16_t lpcomp_setup;
	uint16_t msg_counter;
	float cal_val[4];
	uint32_t config_timestamp;
	uint8_t accel_sensitivity;
	uint8_t accel_freq;
	uint32_t uart_baud;
	
	

	/* CRC used to validate the retained data.  This must be
	 * stored little-endian, and covers everything up to but not
	 * including this field.
	 */
	uint32_t crc;
};

/* For simplicity in the sample just allow anybody to see and
 * manipulate the retained state.
 */
extern struct retained_data retained;

/* Check whether the retained data is valid, and if not reset it.
 *
 * @return true if and only if the data was valid and reflects state
 * from previous sessions.
 */
bool retained_validate(void);

/* Update any generic retained state and recalculate its checksum so
 * subsequent boots can verify the retained state.
 */
void retained_update(void);

#ifdef __cplusplus
}
#endif

#endif /* RETAINED_H_ */
